/*
 * Generic Hash and HMAC Program
 *
 * Copyright (C) 2009 2012 Harald von Fellenberg <hvf@hvf.ch>
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 3 of the License, or 
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* add base64 support - hvf 27.02.2015	*/
/* add xofshakexxx support - hvf 21.04.2015  */
/* shake128, shake256 now support HMAC (NIST SP 800-224) - hvf 31.08.2024 */

#include	"fehashmac.h"
#include	"fehashmac-macros.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <assert.h>
#include "base64.h"

// print base64 output
void	print_b64 (const unsigned char *out, size_t outlen);

// hash one file, multiple hashes, in parallel
// file access uses mmap if possible

void	hash_one_file (const char *filename, struct use_algo *ua)
{
	struct use_algo	*algoptr;
	int	j;
    unsigned char out[HASHBUFSIZ];    /* hash output array */

	int fd;
	FILE *fp;
	if (filename) {	// filename specified
		fp = fopen (filename, "rb");
	} else {		// no filename, use stdin
		fp = stdin;
		filename = "(stdin)";	// fake filename for better output reading
	}
	if (!fp) { 
		printf ("%s: no such file\n", filename);
		perror (filename); 
		return;
	}
	fd = fileno(fp);	// mmap, fstat need file descriptor

	// initialize contexts: work our way through all algorithms 
	for (algoptr = ua; algoptr; algoptr = algoptr->next) {
		struct	hash_algo	*a = algoptr->a;
		HashReturn retval = (*a->init)(&algoptr->context, a->hash_length<<3, 
                            &algoptr->extra);
		if (retval != SUCCESS) {
				fprintf (stderr, 
                "Init context failed for algo %s, reason %d, hash length %d\n",
				a->name, retval, a->hash_length<<3);
                if (strcmp(a->name, "blake3") == 0) {
                    fprintf (stderr, "keyBitLen = %ld, must be %d\n", 
                        algoptr->extra.KeyBitLen, 256);
                }
				exit (1);
			}
		// initialize HMAC if requested
		if (algoptr->hmacflag) {
			HMAC_INIT (algoptr, keyflag, hmacprintflag);
			HMAC_UPDATE (algoptr, algoptr->context);
			hmacprintflag = 0;	// once is enough
		}
		// initialize KMAC if requested
		if (algoptr->kmacflag) {
			KMAC_INIT (algoptr, keyflag, kmacprintflag);
			// KMAC_INIT (algoptr, keyflag, 1);
			KMAC_UPDATE (algoptr, algoptr->context);
			kmacprintflag = 0;	// once is enough
		}
	}

	// now read file in large chunks and process for each algo
	// first try to mmap the open file
	size_t len, filelen = 0;
	int	map_OK = 0;		// set if mmap was successful
	char	*mapaddr = NULL, *endaddr = NULL;

	// try to mmap the file, otherwise read it
	// fstat it first
	struct stat sb;
	if (fstat (fd, &sb)) {	
		perror ("fstat");
	} else {
		if (!S_ISREG(sb.st_mode)) {
		} else {
			if ((filelen = sb.st_size) > 0) {
				mapaddr = mmap (NULL, filelen, PROT_READ, MAP_PRIVATE, fd, 
								(off_t) 0);
				if (mapaddr == MAP_FAILED) {
					perror ("mmap");
				} else {
					map_OK = 1;		// mapping successful
					endaddr = mapaddr + filelen;
				}
			}
		}
	}
			
	// process the mmap'ed file
	if (map_OK) {
		char *fileptr;
		for (fileptr=mapaddr; fileptr<endaddr; fileptr+=LARGEBUFSIZ) {
			len = (fileptr+LARGEBUFSIZ)>endaddr ? endaddr-fileptr : LARGEBUFSIZ;
			// work our way through all algorithms 
			for (algoptr = ua; algoptr; algoptr = algoptr->next) {
				struct	hash_algo	*a = algoptr->a;
				(*a->update)(algoptr->context, 
					(const BitSequence *)fileptr, len<<3);
			}
		}
		munmap (mapaddr, filelen);
	} else {	// read the file
		while ((len = fread (filebuffer, 1, sizeof (filebuffer), fp))) {
			// work our way through all algorithms 
			for (algoptr = ua; algoptr; algoptr = algoptr->next) {
				struct	hash_algo	*a = algoptr->a;
				(*a->update)(algoptr->context, 
					(const BitSequence *)filebuffer, len<<3);
			}
		}
	}
	close (fd);

	// finish processing
	for (algoptr = ua; algoptr; algoptr = algoptr->next) {
		struct	hash_algo	*a = algoptr->a;
		// make sure out is large enough
		assert (sizeof(out) >= a->hash_length);
        // init out to deadbeef - more of a debugging aid
        deadbeef(out, sizeof(out));
        // XXXXXXXXXXX  FIRST CALL TO FINAL  XXXXXXXXXXXXXXXX
		(*a->final)(algoptr->context, out);

		if (algoptr->hmacflag) {
			HMAC_FINAL (algoptr, out);
			printf ("hmac-");
		}
		if (algoptr->kmacflag) {
			// printf ("kmac-");
		}
		// not HMAC, maybe XOF, or base64, or normal
		// all CTX's start with the two identical elements,
		// so the two casts are safe
        // NO!! HMAC now possible for shakennn! 31.08.2024
		GEN_CTX  *gc = (GEN_CTX *) algoptr->context;
		SHA3_CTX *ctx = (SHA3_CTX *) algoptr->context;
        BLAKE3_CTX *b3ctx = (BLAKE3_CTX *) algoptr->context;

        int is_SHA3_type = ((gc->magic == HASH_MAGIC_SHAKE128) || 
                        (gc->magic == HASH_MAGIC_SHAKE256) || 
                        (gc->magic == HASH_MAGIC_KMAC_128) || 
                        (gc->magic == HASH_MAGIC_KMAC_256)) && ctx->xof_OK;
        int is_BLAKE3_type = gc->magic == HASH_MAGIC_BLAKE3 && b3ctx->xof_OK;
		int do_b64 = algoptr->extra.base64flag && a->base64_OK;
        int this_chunk_size = 0; long long more_size = 0;
        if (is_SHA3_type) {
            this_chunk_size = ctx->this_chunk_size;
            more_size = ctx->more_size;
        }
        if (is_BLAKE3_type) {
            this_chunk_size = b3ctx->this_chunk_size;
            more_size = b3ctx->more_size;
        }
        // check for XOF algorithms
		if (is_SHA3_type || is_BLAKE3_type) {
			if (is_SHA3_type) printf ("length (xof) = %lld\n", ctx->xoflength);
			if (is_BLAKE3_type) printf ("length (xof) = %lld\n", b3ctx->xoflength);
            if (algoptr->kmacflag) {
			    printf ("%s%s%s (%s) = ", "kmac-" , 
					a->name, (do_b64 ? "-b64" : ""), filename);
            }
            else if (algoptr->hmacflag) {
			    printf ("%s%s%s (%s) = ", "hmac-" , 
					a->name, (do_b64 ? "-b64" : ""), filename);
            }
            else
			    printf ("%s%s (%s) = ", 
					a->name, (do_b64 ? "-b64" : ""), filename);
			// base64 is possible
			if (do_b64) {
				print_b64 (out, this_chunk_size);
			} else {
                // if (!kmacxofxxx_OK(algoptr->kmacxofxxxflag))
                // initial chunk
				for (j=0; j<this_chunk_size; j++) printf ("%2.2x", out[j]&0xff);
			}
			// XXXXXXXXXXX SECOND AND LATER CHUNKS - CALLS TO FINAL
			while ((is_SHA3_type && (ctx->more_size && ctx->xofflag)) ||
                (is_BLAKE3_type && (b3ctx->more_size && b3ctx->xofflag))) {	

                if (is_SHA3_type) {
                    this_chunk_size = ctx->this_chunk_size;
                    more_size = ctx->more_size;
                }
                if (is_BLAKE3_type) {
                    this_chunk_size = b3ctx->this_chunk_size;
                    more_size = b3ctx->more_size;
                }
				(*a->final)(algoptr->context, out);

                if (is_SHA3_type) {
                    this_chunk_size = ctx->this_chunk_size;
                    more_size = ctx->more_size;
                }
                if (is_BLAKE3_type) {
                    this_chunk_size = b3ctx->this_chunk_size;
                    more_size = b3ctx->more_size;
                }
				if (do_b64) {
					print_b64 (out, this_chunk_size);
				} else {
					for (j=0; j<this_chunk_size; j++) printf ("%2.2x", out[j]&0xff);
				}
			}
			putchar ('\n');
		}
		else {  // no xof algorithm, normal output
			printf ("%s%s%s (%s) = ", (algoptr->kmacflag ? "kmac-" : ""),
					a->name, (do_b64 ? "-b64" : ""), filename);
			// base64 is possible
			if (do_b64) {
				print_b64 (out, a->hash_length);
			} else {
				for (j=0; j<a->hash_length; j++) printf ("%2.2x", out[j]&0xff);
			}
			putchar ('\n');
		}
		(*a->free)(algoptr->context); algoptr->context = NULL;
	}
	
	return;
}

// print base64 output
void	print_b64 (const unsigned char *out, size_t outlen)
{
	char *b64out;
	size_t b64outlen = base64_encode_alloc ((const char *) out, 
						outlen, &b64out);
	if (b64out == NULL && b64outlen == 0 && outlen != 0) {
		fprintf (stderr, "base64 input too long.\n");
		exit (1);
	}
	if (b64out == NULL) {
		fprintf (stderr, "base64  memory allocation error.\n");
			exit (1);
	}
	int j;
	for (j=0; j<b64outlen; j++) putchar (b64out[j]);
	free (b64out);
}
